import re
from xml.etree.ElementTree import Element

from Dictionary.Word import Word
from MorphologicalAnalysis.MorphologicalParse import MorphologicalParse

from DependencyParser.Turkish.TurkishDependencyRelation import TurkishDependencyRelation


class TurkishDependencyTreeBankWord(Word):

    __parse: MorphologicalParse
    __original_parses: list
    __relation: TurkishDependencyRelation

    def __init__(self, wordNode: Element):
        """
        Given the parsed xml node which contains information about a word and related attributes including the
        dependencies, the method constructs a {@link TurkishDependencyTreeBankWord} from it.

        PARAMETERS
        -----------
        wordNode : Element
            Xml parsed node containing information about a word.
        """
        to_word = 0
        to_ig = 0
        self.__original_parses = []
        self.name = wordNode.text
        ig = wordNode.attrib["IG"]
        ig = ig[:ig.index("+")] + "+" + ig[ig.index("+") + 1:].upper()
        self.__parse = MorphologicalParse(self.splitIntoInflectionalGroups(ig))
        self.__relation = None
        relation_name = wordNode.attrib["REL"]
        if relation_name != "[,( )]":
            relation_parts = re.compile("[\\[()\\],]").split(relation_name)
            index = 0
            for part in relation_parts:
                if len(part) != 0:
                    index = index + 1
                    if index == 1:
                        to_word = int(part)
                    elif index == 2:
                        to_ig = int(part)
                    elif index == 3:
                        self.__relation = TurkishDependencyRelation(to_word - 1, to_ig - 1, part)
        for i in range(1, 10):
            if ("ORG_ID" + str(i)) in wordNode.attrib:
                ig = wordNode.attrib["ORG_ID" + str(i)]
                ig = ig[:ig.index("+")] + "+" + ig[ig.index("+") + 1:].upper()
                self.__original_parses.append(MorphologicalParse(self.splitIntoInflectionalGroups(ig)))

    def splitIntoInflectionalGroups(self, IG: str) -> list:
        """
        Given the morphological parse of a word, this method splits it into inflectional groups.

        PARAMETERS
        ----------
        IG : str
            Morphological parse of the word in string form.

        RETURNS
        -------
        list
            A list of inflectional groups stored as strings.
        """
        inflectional_groups = []
        IG = IG.replace("(+Punc", "@").replace(")+Punc", "$")
        IGs = re.compile("[\\[()\\]]").split(IG)
        for part in IGs:
            part = part.replace("@", "(+Punc").replace("$", ")+Punc")
            if len(part) != 0:
                inflectional_groups.append(part)
        return inflectional_groups

    def getParse(self) -> MorphologicalParse:
        """
        Accessor for the parse attribute

        RETURNS
        -------
        MorphologicalPArse
            Parse attribute
        """
        return self.__parse

    def getOriginalParse(self, index: int) -> MorphologicalParse:
        """
        Accessor for a specific parse.

        PARAMETERS
        ----------
        index : int
            Index of the word.

        RETURNS
        -------
        MorphologicalParse
            Parse of the index'th word
        """
        if index < len(self.__original_parses):
            return self.__original_parses[index]
        else:
            return None

    def size(self) -> int:
        """
        Number of words in this item.

        RETURNS
        -------
        int
            Number of words in this item.
        """
        return len(self.__original_parses)

    def getRelation(self) -> TurkishDependencyRelation:
        """
        Accessor for the relation attribute.

        RETURNS
        -------
        TurkishDependencyRelation
            relation attribute.
        """
        return self.__relation

    def __repr__(self):
        return f"{self.name} {self.__parse} {self.__original_parses} {self.__relation}"



from __future__ import annotations


class ParserEvaluationScore:

    LAS: float
    UAS: float
    LS: float
    word_count: int

    def __init__(self, LAS: float = 0.0, UAS: float = 0.0, LS: float = 0.0, wordCount: int = 0):
        self.LAS = LAS
        self.UAS = UAS
        self.LS = LS
        self.word_count = wordCount

    def getLS(self) -> float:
        return self.LS

    def getLAS(self) -> float:
        return self.LAS

    def getUAS(self) -> float:
        return self.UAS

    def getWordCount(self) -> int:
        return self.word_count

    def add(self, parserEvaluationScore: ParserEvaluationScore):
        self.LAS = (self.LAS * self.word_count + parserEvaluationScore.LAS * parserEvaluationScore.word_count) / \
                   (self.word_count + parserEvaluationScore.word_count)
        self.UAS = (self.UAS * self.word_count + parserEvaluationScore.UAS * parserEvaluationScore.word_count) / \
                   (self.word_count + parserEvaluationScore.word_count)
        self.LS = (self.LS * self.word_count + parserEvaluationScore.LS * parserEvaluationScore.word_count) / \
                  (self.word_count + parserEvaluationScore.word_count)
        self.word_count += parserEvaluationScore.word_count



class UniversalDependencyTreeBankFeatures:

    feature_list: dict

    def __init__(self, features: str):
        self.feature_list = {}
        if features != "_":
            _list = features.split("\\|")
            for feature in _list:
                if "=" in feature:
                    feature_name = feature[0: feature.index("=") - 1].strip()
                    feature_value = feature[feature.index("=") + 1:].strip()
                    self.feature_list[feature_name] = feature_value
                else:
                    print("Feature does not contain = ->" + features)

    def getFeatureValue(self, feature: str) -> str:
        return self.feature_list[feature]

    def featureExists(self, feature: str) -> bool:
        return feature in self.feature_list

    def __str__(self) -> str:
        if len(self.feature_list) == 0:
            return "_"
        result = ""
        for feature in self.feature_list:
            if result == "":
                result = feature + "=" + self.getFeatureValue(feature)
            else:
                result += "|" + feature + "=" + self.getFeatureValue(feature)
        return result



from __future__ import annotations

from DependencyParser.DependencyRelation import DependencyRelation
from DependencyParser.ParserEvaluationScore import ParserEvaluationScore
from DependencyParser.Universal.UniversalDependencyPosType import UniversalDependencyPosType
from DependencyParser.Universal.UniversalDependencyType import UniversalDependencyType


class UniversalDependencyRelation(DependencyRelation):
    __universal_dependency_type: UniversalDependencyType

    universal_dependency_types = ["ACL", "ADVCL",
                                  "ADVMOD", "AMOD",
                                  "APPOS",
                                  "AUX", "AUXPASS",
                                  "CASE", "CC", "CCOMP",
                                  "CLF", "COMPOUND",
                                  "CONJ", "COP", "CSUBJ",
                                  "DEP", "DET",
                                  "DISCOURSE", "DISLOCATED",
                                  "EXPL",
                                  "FIXED", "FLAT",
                                  "FOREIGN",
                                  "GOESWITH",
                                  "IOBJ",
                                  "LIST",
                                  "MARK",
                                  "NMOD", "NSUBJ",
                                  "NUMMOD",
                                  "OBJ", "OBL", "ORPHAN",
                                  "PARATAXIS", "PUNCT",
                                  "REPARANDUM", "ROOT",
                                  "VOCATIVE",
                                  "XCOMP", "ACL:RELCL",
                                  "AUX:PASS", "CC:PRECONJ",
                                  "COMPOUND:PRT", "DET:PREDET",
                                  "FLAT:FOREIGN", "NSUBJ:PASS",
                                  "CSUBJ:PASS", "NMOD:NPMOD",
                                  "NMOD:POSS", "NMOD:TMOD",
                                  "ADVMOD:EMPH", "AUX:Q",
                                  "COMPOUND:LVC", "COMPOUND:REDUP",
                                  "CSUBJ:COP", "NMOD:COMP",
                                  "NMOD:PART", "NSUBJ:COP",
                                  "OBL:AGENT", "OBL:TMOD",
                                  "OBL:NPMOD"]

    universal_dependency_tags = [UniversalDependencyType.ACL, UniversalDependencyType.ADVCL,
                                 UniversalDependencyType.ADVMOD, UniversalDependencyType.AMOD,
                                 UniversalDependencyType.APPOS,
                                 UniversalDependencyType.AUX, UniversalDependencyType.AUXPASS,
                                 UniversalDependencyType.CASE, UniversalDependencyType.CC,
                                 UniversalDependencyType.CCOMP,
                                 UniversalDependencyType.CLF, UniversalDependencyType.COMPOUND,
                                 UniversalDependencyType.CONJ, UniversalDependencyType.COP,
                                 UniversalDependencyType.CSUBJ,
                                 UniversalDependencyType.DEP, UniversalDependencyType.DET,
                                 UniversalDependencyType.DISCOURSE, UniversalDependencyType.DISLOCATED,
                                 UniversalDependencyType.EXPL,
                                 UniversalDependencyType.FIXED, UniversalDependencyType.FLAT,
                                 UniversalDependencyType.FOREIGN,
                                 UniversalDependencyType.GOESWITH,
                                 UniversalDependencyType.IOBJ,
                                 UniversalDependencyType.LIST,
                                 UniversalDependencyType.MARK,
                                 UniversalDependencyType.NMOD, UniversalDependencyType.NSUBJ,
                                 UniversalDependencyType.NUMMOD,
                                 UniversalDependencyType.OBJ, UniversalDependencyType.OBL,
                                 UniversalDependencyType.ORPHAN,
                                 UniversalDependencyType.PARATAXIS, UniversalDependencyType.PUNCT,
                                 UniversalDependencyType.REPARANDUM, UniversalDependencyType.ROOT,
                                 UniversalDependencyType.VOCATIVE,
                                 UniversalDependencyType.XCOMP, UniversalDependencyType.ACL_RELCL,
                                 UniversalDependencyType.AUX_PASS, UniversalDependencyType.CC_PRECONJ,
                                 UniversalDependencyType.COMPOUND_PRT,
                                 UniversalDependencyType.DET_PREDET, UniversalDependencyType.FLAT_FOREIGN,
                                 UniversalDependencyType.NSUBJ_PASS, UniversalDependencyType.CSUBJ_PASS,
                                 UniversalDependencyType.NMOD_NPMOD, UniversalDependencyType.NMOD_POSS,
                                 UniversalDependencyType.NMOD_TMOD, UniversalDependencyType.ADVMOD_EMPH,
                                 UniversalDependencyType.AUX_Q, UniversalDependencyType.COMPOUND_LVC,
                                 UniversalDependencyType.COMPOUND_REDUP, UniversalDependencyType.CSUBJ_COP,
                                 UniversalDependencyType.NMOD_COMP, UniversalDependencyType.NMOD_PART,
                                 UniversalDependencyType.NSUBJ_COP, UniversalDependencyType.OBL_AGENT,
                                 UniversalDependencyType.OBL_TMOD, UniversalDependencyType.OBL_NPMOD]

    universal_dependency_pos_types = ["ADJ", "ADV", "INTJ", "NOUN", "PROPN", "VERB", "ADP", "AUX", "CCONJ",
                                      "DET", "NUM", "PART", "PRON", "SCONJ", "PUNCT", "SYM", "X"]

    universal_dependency_pos_tags = [UniversalDependencyPosType.ADJ, UniversalDependencyPosType.ADV,
                                     UniversalDependencyPosType.INTJ, UniversalDependencyPosType.NOUN,
                                     UniversalDependencyPosType.PROPN,
                                     UniversalDependencyPosType.VERB, UniversalDependencyPosType.ADP,
                                     UniversalDependencyPosType.AUX, UniversalDependencyPosType.CCONJ,
                                     UniversalDependencyPosType.DET, UniversalDependencyPosType.NUM,
                                     UniversalDependencyPosType.PART,
                                     UniversalDependencyPosType.PRON, UniversalDependencyPosType.SCONJ,
                                     UniversalDependencyPosType.PUNCT, UniversalDependencyPosType.SYM,
                                     UniversalDependencyPosType.X]

    @staticmethod
    def getDependencyTag(tag: str) -> UniversalDependencyType:
        """
        The getDependencyTag method takes an dependency tag as string and returns the UniversalDependencyType
        form of it.

        PARAMETERS
        ----------
        tag : str
            Type of the dependency tag in string form

        RETURNS
        -------
        UniversalDependencyType
            Type of the dependency in UniversalDependencyType form
        """
        for j in range(len(UniversalDependencyRelation.universal_dependency_tags)):
            if tag.upper() == UniversalDependencyRelation.universal_dependency_types[j]:
                return UniversalDependencyRelation.universal_dependency_tags[j]
        return None

    @staticmethod
    def getDependencyPosType(tag: str) -> UniversalDependencyPosType:
        for j in range(len(UniversalDependencyRelation.universal_dependency_pos_types)):
            if tag.upper() == UniversalDependencyRelation.universal_dependency_pos_types[j]:
                return UniversalDependencyRelation.universal_dependency_pos_tags[j]
        return None

    def __init__(self,
                 toWord: int,
                 dependencyType: str = None):
        """
        Overriden Universal Dependency Relation constructor. Gets toWord as input and calls it super class's constructor

        PARAMETERS
        ----------
        toWord : int
            Index of the word in the sentence that dependency relation is related
        """
        super().__init__(toWord)
        if dependencyType is not None:
            self.__universal_dependency_type = UniversalDependencyRelation.getDependencyTag(dependencyType)

    def compareRelations(self, relation: UniversalDependencyRelation) -> ParserEvaluationScore:
        LS = 0.0
        LAS = 0.0
        UAS = 0.0
        if self.__str__() == relation.__str__():
            LS = 1.0
            if self.to_word == relation.to():
                LAS = 1.0
        if self.to_word == relation.to():
            UAS = 1.0
        return ParserEvaluationScore(LAS, UAS, LS, 1)

    def __str__(self) -> str:
        return self.__universal_dependency_type.name
